////////////////////////////////////////////////////////////////////////////////
//
//  ****************************************************************************
//  * Unit Name : SimpleRoundToEx
//  * Purpose   :   .
//  * Author    :  (Rouse_) 
//  * Copyright :  Fangorn Wizards Lab 1998 - 2012.
//  * Version   : 1.0.0
//  * Home Page : http://rouse.drkb.ru
//  ****************************************************************************
//

unit SimpleRoundToEx;

interface

uses
  SysUtils,
  Math;

  function SimpleRoundTo_Asm(const AValue: Extended; const ADigit: TRoundToRange = -2): Extended;
  function SimpleRoundTo_Str(const AValue: Extended; ADigit: TRoundToRange = -2): Extended;


implementation

const
  // 8087 status word masks
  mC0 = $0100;
  mC2 = $0400;
  mC3 = $4000;

  Multiply: array [0..29] of Double = (
    1E01, 1E02, 1E03, 1E04, 1E05, 1E06, 1E07, 1E08, 1E09,
    1E10, 1E11, 1E12, 1E13, 1E14, 1E15, 1E16, 1E17, 1E18, 1E19,
    1E20, 1E21, 1E22, 1E23, 1E24, 1E25, 1E26, 1E27, 1E28, 1E29, 1E30);
  Divide: array [0..29] of Double =
    (1, 1E-01, 1E-02, 1E-03, 1E-04, 1E-05, 1E-06, 1E-07, 1E-08, 1E-09,
    1E-10, 1E-11, 1E-12, 1E-13, 1E-14, 1E-15, 1E-16, 1E-17, 1E-18, 1E-19,
    1E-20, 1E-21, 1E-22, 1E-23, 1E-24, 1E-25, 1E-26, 1E-27, 1E-28, 1E-29);

procedure RaiseInvalidOP;
begin
  raise EInvalidOp.Create('Invalid floating point operation');
end;

procedure NormalizeAndTruncExtended;
var
  ExtValue: Extended;
  StatWord: Word;
asm
        jmp @start

  @fpu_error:
        call RaiseInvalidOP

        //   
        //      FFFX  
        //     
        //   FFFF

  @start:
        pusha
        xor edx, edx
        fld st(0)
        lea eax, ExtValue
        fstp tbyte ptr[eax]
        movzx ecx, word ptr [eax]
        cmp ecx, $FFF0
        jl @value_normalized

  @normalize_loop:
        mov word ptr [eax], 0
        add eax, 2
        add edx, 2
        movzx ecx, word ptr [eax]
        inc ecx
        mov word ptr [eax], cx
        cmp cx, 0
        jne @load_new_value
        jmp @normalize_loop

  @load_new_value:
        sub eax, edx
        fld tbyte ptr[eax]
        fxch st(1)
        fstp tbyte ptr[eax]

        fxam
        fstsw StatWord
        and StatWord, mC2
        cmp StatWord, mC2
        je @value_normalized

        //    -   c  
        //     ,        
        fstp st(0)
        fld tbyte ptr[eax]
        fld1
        fld st(1)
        fprem1
        fstsw   ax
        sahf
        jp @fpu_error

        fadd st(2), st(0)
        jmp @done

  @value_normalized:

        // Result := Trunc(Result)
        //   FRNDINT
        //        -  
        fld1
        fld st(1)
        fprem
        fstsw   ax
        sahf
        jp @fpu_error

        fsub st(2), st(0)

        //  
  @done:
        fstp st(0)
        fstp st(0)
        popa
end;

//         BCD
function ImproveAccuracy(const Value: Extended): Extended;
const
  E18: Extended = 1E18;
  I10: Integer = 10;
var
  BCDValue: Extended;
  StatWord: Word;
  Signed: Boolean;
asm
        //   
        fldz
        fld tbyte ptr [ebp + 8]
        fcompp
        fstsw StatWord
        movzx eax, StatWord
        and eax, mC3
        cmp eax, mC3
        je @zero
        and StatWord, mC0
        xor eax, eax
        cmp StatWord, mC0
        sete al
        mov Signed, al

        //   
  @not_signed:
        fld tbyte ptr [ebp + 8]
        lea eax, BCDValue
        fstp tbyte ptr [eax]
        mov ax, [eax].Word[8]
        and eax, $7FFF
        sub eax, $3FFF
        imul eax, 19728
        sar eax, 16
        inc eax
        mov ecx, eax
        mov eax, 18
        sub eax, ecx
        fld1

        //     
        push ebx
        call FPower10
        pop ebx
        fld tbyte ptr [ebp + 8]
        fmul st(0), st(1)

        // 
        frndint
        fabs

        //        E18  ( BCD )
        lea eax, E18
        fld tbyte ptr [eax]
        fcomp
        fstsw StatWord
        test StatWord, mC0+mC3
        je @store_bcd
        lea eax, I10
        fild [eax]
        fdiv st(1), st(0)
        fdivp st(2), st(0)

  @store_bcd:
        //  BCD 
        fbstp BCDValue

        lea eax, BCDValue
        mov byte ptr [eax], 0
        mov cx, 5

        //      FloatToStr
  @increment_next_word:
        inc eax
        mov dl, byte ptr [eax]
        add dl, ch
        mov dh, dl
        shr dl, 4
        and dh, $F

        cmp dh, cl
        jl @done

        mov dh, 0
        inc dl

        cmp dl, 10
        jl  @done

        mov byte ptr [eax], 0

        mov cx, $10A

        jmp @increment_next_word

  @done:

        //  
        shl dl, 4
        add dl, dh
        mov byte ptr [eax], dl

        fbld BCDValue
        fdiv st(0), st(1)
        fstp st(1)

        //  
        movzx eax, Signed
        cmp eax, 0
        je @exit
        fchs
        jmp @exit

  @zero:
        fldz

  @exit:
end;

function _SimpleRoundTo(const AValue: Extended;
  const ADigit: TRoundToRange = -2): Extended;
const
  IncValue: Extended = 0.5;
asm
        // LFactor := IntPower(10, ADigit);
        push $00004002
        push $a0000000
        push $00
        movsx eax, al
        call IntPower

        //  
        // DH := 0;
        // if Value < 0 then
        // begin
        //   Value := Value * -1;
        //   DH := 1;
        // end;
        xor edx, edx
        fld tbyte ptr [ebp + 8]
        fldz
        fcomp
        fstsw   ax
        sahf
        jbe @positive
        fchs
        mov dh, 1

  @positive:
        // Result := (AValue / LFactor) + 0.5
        fdiv st(0), st(1)
        lea eax, IncValue
        fld tbyte ptr[eax]
        faddp st(1), st(0)

        call NormalizeAndTruncExtended

        // Result := Result * LFactor
        fmulp st(1), st(0)

        //  
        // if DH = 1 then Value := Value * -1;
        cmp dh, 0
        jz @exit
        fchs

  @exit:
end;

function SimpleRoundTo_Str(const AValue: Extended;
  ADigit: TRoundToRange = -2): Extended;
var
  I, DotIndex, InitialLength, LastCharPos, CopyLength: Integer;
  LastChar: Char;
  NeedIncrement: Boolean;
  StrValue, InitialValue: string;
begin

  if AValue < 0 then
    InitialValue := FloatToStrF(-AValue, ffFixed, 15, 35)
  else
    InitialValue := FloatToStrF(AValue, ffFixed, 15, 35);

  LastChar := #0;
  InitialLength := Length(InitialValue);
  DotIndex := Pos(DecimalSeparator, InitialValue);

  LastCharPos := 0;
  CopyLength := 0;
  if DotIndex = 0 then
  begin
    if ADigit <= 0 then
      CopyLength := InitialLength
    else
    begin
      if ADigit < InitialLength then
        LastCharPos := InitialLength - ADigit + 1;
      CopyLength := Max(0, InitialLength - ADigit);
    end;
  end
  else
  begin
    if ADigit < 0 then
    begin
      if DotIndex - ADigit < InitialLength then
        LastCharPos := DotIndex - ADigit + 1;
      CopyLength := Min(DotIndex - ADigit, InitialLength);
    end;
    if ADigit = 0 then
    begin
      LastCharPos := DotIndex + 1;
      CopyLength := DotIndex - 1;
    end;
    if ADigit > 0 then
    begin
      if DotIndex - 1 = ADigit then
        LastCharPos := 1;
      if DotIndex > ADigit then
      begin
        LastCharPos := DotIndex - ADigit;
        CopyLength := Max(0, DotIndex - ADigit - 1);
      end;
    end;
  end;

  if LastCharPos <> 0 then
    LastChar := InitialValue[LastCharPos];

  if CopyLength > 0 then
    StrValue := Copy(InitialValue, 1, CopyLength)
  else
    StrValue := '0';

  if LastChar <> #0 then
  begin
    I := Length(StrValue);
    if I = 0 then StrValue := '0';
    if StrValue[I] = DecimalSeparator then
    begin
      StrValue := Copy(StrValue, 1, I - 1);
      Dec(I);
    end;
    if CharInSet(LastChar, ['5'..'9']) then
    begin
      NeedIncrement := True;
      while NeedIncrement and (I > 0) do
      begin
        if StrValue[I] = DecimalSeparator then
        begin
          Dec(I);
          Continue;
        end;
        if StrValue[I] = '9' then
        begin
          StrValue[I] := '0';
          Dec(I);
          if I = 0 then
            StrValue := '1' + StrValue
          else
            if (I = 1) and (StrValue[1] = '-') then
            begin
              StrValue[1] := '1';
              StrValue := '-' + StrValue;
              I := 0;
            end;
        end
        else
        begin
          StrValue[I] := Char(Byte(StrValue[I]) + 1);
          NeedIncrement := False;
        end;
      end;
    end;
  end;

  if ADigit >= 0 then
  begin
    while ADigit <> 0 do
    begin
      Dec(ADigit);
      StrValue := StrValue + '0';
    end;
  end;

  if StrValue <> '0' then
    if AValue < 0 then
      StrValue := '-' + StrValue;

  Result := StrToFloat(StrValue);
end;

function SimpleRoundTo_Asm(const AValue: Extended;
  const ADigit: TRoundToRange = -2): Extended;
var
  ConvertedValue: Extended;
begin
  ConvertedValue := ImproveAccuracy(AValue);
  if ConvertedValue = 0 then
    ConvertedValue := AValue;
  Result := _SimpleRoundTo(ConvertedValue, ADigit);
end;
end.
